"""
File: hashtable.py

Case study for Chapter 19.
"""

from arrays import Array

class HashTable(object):
    "Represents a hash table."""

    EMPTY = None
    DELETED = True

    def __init__(self, capacity = 29,
                 hashFunction = hash,
                 linear = True):
        self._table = Array(capacity, HashTable.EMPTY)
        self._size = 0
        self._hash = hashFunction
        self._homeIndex = -1
        self._actualIndex = -1
        self._linear = linear
        self._probeCount = 0

    def insert(self, item):
        """Inserts item into the table
        Preconditions: There is at least one empty cell or
        one previously occupied cell.
        There is not a duplicate item."""
        self._probeCount = 0
        # Get the home index
        self._homeIndex = abs(self._hash(item)) % len(self._table)
        distance = 1
        index = self._homeIndex

        # Stop searching when an empty cell is encountered
        while not self._table[index] in (HashTable.EMPTY,
                                         HashTable.DELETED):

            # Increment the index and wrap around to first 
            # position if necessary
            if self._linear:
                increment = index + 1
            else:
                # Quadratic probing
                increment = self._homeIndex + distance ** 2
                distance += 1
            index = increment % len(self._table)
            self._probeCount += 1

        # An empty cell is found, so store the item
        self._table[index] = item
        self._size += 1
        self._actualIndex = index

    def search(self, item):
        """Search for item in the table."""
        self._probeCount = 0
        # Get the home index
        self._homeIndex = abs(self._hash(item)) % len(self._table)
        distance = 1
        index = self._homeIndex

        # Stop searching when an empty cell is encountered
        while not self._table[index] in (HashTable.EMPTY,
                                         item):

            # Increment the index and wrap around to first 
            # position if necessary
            if self._linear:
                increment = index + 1
            else:
                # Quadratic probing
                increment = self._homeIndex + distance ** 2
                distance += 1
            index = increment % len(self._table)
            self._probeCount += 1

        # An empty cell is found, so return None
        if self._table[index] == HashTable.EMPTY:
            return None
        else:
            self._actualIndex = index
            return self._table[index]

    # Methods __len__(), __str__(), loadFactor(), homeIndex(),
    # actualIndex(), and probeCount() are exercises.
    def __len__(self):
        return self._size

    def __str__(self):
        return str(self._table)

    def loadFactor(self):
        return float(self._size)/len(self._table)

    def homeIndex(self):
        return self._homeIndex

    def actualIndex(self):
        return self._actualIndex

    def probeCount(self):
        return self._probeCount


    
        

       
def main():
    """Uses an example data set from Chapter 19."""
    table = HashTable(8, lambda x : x)
    for item in (range(10, 71, 10)):
        table.insert(item)
        print "Home:", table.homeIndex(), "Probes:", table.probeCount(), \
              "Load factor:", table.loadFactor()
        print table

if __name__ == "__main__": main()

        



